from logging import Logger
from typing import Any, Dict, Union, OrderedDict, Protocol
from inginious.frontend.user_manager import UserManager

# An INGInious submission
Submission = OrderedDict[str, Any]

# Grades from a single category
GradingCategoryIn = Dict[str, Union[str, int]]

# Mapping of coding style grades
GradesIn = Dict[str, GradingCategoryIn]


class INGIniousPageProto(Protocol):
    """Base class for INGInious page protocol types.
    Only stipulates that a logger must be present for now."""

    _logger: Logger


class HasUserManager(INGIniousPageProto):
    """Protocol type that stipulates the object must have access to the user manager."""

    @property
    def user_manager(self) -> UserManager:
        ...



import time
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Union, TYPE_CHECKING

from bson.errors import InvalidId
from flask import redirect, request
from inginious.client.client import Client
from inginious.frontend.course_factory import CourseFactory
from inginious.frontend.courses import Course
from inginious.frontend.pages.course_admin.submission import SubmissionPage
from inginious.frontend.pages.utils import INGIniousAuthPage
from inginious.frontend.plugin_manager import PluginManager
from inginious.frontend.tasks import Task
from inginious.frontend.template_helper import TemplateHelper
from pydantic import ValidationError
from pymongo.collection import ReturnDocument
from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.exceptions import BadRequest, Forbidden, InternalServerError, NotFound
from werkzeug.wrappers.response import Response

from ._types import GradesIn, Submission
from .config import PluginConfig, get_config
from .grades import CodingStyleGrades, get_grades
from .logger import get_logger
from .utils import (
    get_submission_authors_realname,
    get_submission_timestamp,
    has_coding_style_grades,
    get_best_submission,
)

if TYPE_CHECKING:
    from datetime import datetime

__version__ = "1.0.1"

PLUGIN_PATH = Path(__file__).parent.absolute()
TEMPLATES_PATH = PLUGIN_PATH / "templates"

# The key to store style grade data in the submission's "custom" dict
PLUGIN_KEY = "coding_style"  # NOTE: use plugin name from config?

# Makes plugin config available globally
PLUGIN_CONFIG: PluginConfig = None  # type: ignore


class StudentSubmissionCodingStyle(INGIniousAuthPage):
    """Displays a detailed view of coding style grades for a single submission
    made by a student."""

    _logger = get_logger()

    def __init__(self, config: PluginConfig, *args, **kwargs) -> None:
        self.config = config
        super().__init__(*args, **kwargs)

    def GET_AUTH(self, submissionid: str) -> str:
        """Displays all coding style grades for a given course for a user."""
        submission = self.get_submission(submissionid)
        course = self.get_course(submission)
        task = self.get_task(submission, course)

        if not has_coding_style_grades(submission, PLUGIN_KEY):
            raise NotFound("Submission has no coding style grades.")

        try:
            grades = get_grades(submission["custom"][PLUGIN_KEY])
        except ValidationError:
            msg = (
                f"Unable to process coding style grades for submission {submissionid}."
            )
            self._logger.exception(msg)
            raise InternalServerError(msg)

        names = get_submission_authors_realname(self, submission)
        submitted_on = get_submission_timestamp(submission)

        return self.template_helper.render(
            "stylegrade.html",
            template_folder=TEMPLATES_PATH,
            submission_authors=names,
            submitted_on=submitted_on,
            user_manager=self.user_manager,
            course=course,
            task=task,
            submission=submission,
            grades=grades,
            config=self.config,
        )

    def get_course(self, submission: Submission) -> Course:
        try:
            courseid = submission.get("courseid")
            course = self.course_factory.get_course(courseid)
        except Exception as e:
            if not courseid:
                self._logger.error(
                    f"Submission {submission['_id']} is not associated with any course. Has the submission been corrupted?"
                )
            else:
                self._logger.error(
                    f"Unable to find course with course ID {courseid}. Has it been deleted?"
                )
            raise InternalServerError("Unable to display submission.")
        return course

    def get_task(self, submission: Submission, course: Course) -> Task:
        try:
            taskid = submission.get("taskid")
            task = course.get_task(taskid)
        except Exception as e:
            if not taskid:
                self._logger.error(
                    f"Submission {submission['_id']} is not associated with a task. Has the submission been corrupted?"
                )
            else:
                self._logger.error(
                    f"Unable to find task with task ID {taskid}. Has it been deleted?"
                )
            raise InternalServerError("Unable to display submission.")
        return task

    def get_submission(self, submissionid: str) -> Submission:
        """Slimmed down version of SubmissionPage.fetch_submission.
        Only returns Submission, instead of Tuple[Course, Task, Submission]"""

        try:
            submission = self.submission_manager.get_submission(submissionid, False)
            if not submission:
                raise NotFound(description=_("This submission doesn't exist."))
        except InvalidId as ex:
            self._logger.info("Invalid ObjectId : %s", submissionid)
            raise Forbidden(description=_("Invalid ObjectId."))
        return submission


class CodingStyleGrading(SubmissionPage):
    """Page that lets administrators grade the coding style of a submission."""

    _logger = get_logger()

    def __init__(self, config: PluginConfig, *args, **kwargs) -> None:
        self.config = config
        super().__init__(*args, **kwargs)

    def GET_AUTH(self, submissionid: str) -> str:
        """Displays coding style grading page for a specific submission."""
        course, task, submission = self.fetch_submission(submissionid)

        # Ensure the submission has a CodingStyleGrades object at ["custom"][PLUGIN_KEY].
        # Loads existing style grades or creates new.

        # TODO: use has_coding_style_grades()?
        submission = self.ensure_submission_custom_key(submission)
        grades = submission["custom"][PLUGIN_KEY]  # type: CodingStyleGrades

        authors = get_submission_authors_realname(self, submission)
        submitted_on = get_submission_timestamp(submission)

        # Check if page is displayed after updating submission grades
        # Display alert denoting success of update:
        # None = no msg, True = success msg, False = failure msg
        success = request.args.get("success")

        return self.template_helper.render(
            "grade_submission.html",
            template_folder=TEMPLATES_PATH,
            user_manager=self.user_manager,
            submitted_on=submitted_on,
            authors=authors,
            course=course,
            task=task,
            submission=submission,
            grades=grades,
            config=self.config,
            success=success,
        )

    def POST_AUTH(self, submissionid: str) -> Response:
        """Adds or updates the coding style grades of a submission."""
        # Parse grades from grading form
        grades = self.parse_form_data(request.form)

        success = True
        try:
            self.update_submission_grades(submissionid, grades)
        except Exception as e:
            self._logger.exception(
                f"Failed to validate request body for submission {submissionid}: {grades}"
            )
            success = False

        # Redirect to updated submission
        return redirect(
            f"/admin/codingstyle/submission/{submissionid}?success={1 if success else 0}"
        )

    def put(self, submissionid: str, *args, **kwargs) -> Union[str, Response]:
        """We (ab)use the superclass `flask.views.MethodView` here to add a PUT rule for the view.

        INGInious only implements their own `GET_AUTH` and `POST_AUTH` methods,
        so we use this method to add support for partial updates of Coding Style Grades,
        such as updating only a single category or removing a category altogether.

        Right now, the only supported operation is removing a grading category
        from a submission.
        """

        # Retrieve submission, then check permissions.
        submission = self.submission_manager.get_submission(
            submissionid, user_check=False
        )
        self.get_course_and_check_rights(submission["courseid"], submission["taskid"])

        # Check if a category should be removed
        if category := request.args.get("remove"):
            self.remove_category_from_submission(submission, category=category)
            return Response(
                "ok",
                # Use htmx custom htmx header to prompt browser to refresh page
                # Source: https://htmx.org/docs/#response-headers
                headers={"HX-Refresh": "true"},
            )
        # TODO: add other PUT operations
        raise BadRequest("Unsupported operation")

    def remove_category_from_submission(
        self, submission: Submission, category: str
    ) -> None:
        """Removes a category from a submission.
        Does nothing if submission does not have a category with that name.

        Parameters
        ----------
        submission : Submission
            The submission to modify.
        category : str
            Name of the category to remove.
        """
        submission = self.ensure_submission_custom_key(submission)
        submission["custom"][PLUGIN_KEY].remove_category(category)
        self._update_submission(submission)

    def parse_form_data(self, form_data: ImmutableMultiDict) -> GradesIn:
        """Transforms flat form data into nested data that can be parsed
        by `CodingStyleGrades.parse_obj()`

        ### Example:

        >>> form_data.to_dict()
        {
            "comments_grade": "100",
            "comments_feedback": "Very good!"
            "modularity_grade": "50",
            "modularity_feedback": "Needs more functions!"
        }
        >>> parse_form_data(form_data)
        {
            "comments": {
                "grade": "100",
                "feedback": "Very good!",
            },
            "modularity": {
                "grade": "50",
                "feedback": "Needs more functions!",
            }
        }
        """
        form = form_data.to_dict()  # type: Dict[str, str]

        out: GradesIn = {}
        for (k, v) in form.items():
            category, attr = k.split("_")
            try:
                out[category][attr] = v
            except KeyError:
                out[category] = {attr: v}
        return out

    def ensure_submission_custom_key(self, submission: Submission) -> Submission:
        """Adds or instantiates data structure for custom grades to a submission.

        Stores existing custom submission data (if it exists) under `submission["custom"]["original"]`,
        if existing custom data is not a dict.

        Updates the submission's database record if coding style grades datastructure
        is added to the submission.

        ### Rationale:

        There is no formal interface for adding custom data to a submission,
        or indeed any specific structure that should be adhered to, so essentially any plugin can
        add whatever they want to the "custom" key as long as it is serializable,
        and can as a result overwrite other plugins' custom submission data with impunity.

        This method makes sure the value for the "custom" key of an INGInious submission is a dict.
        By default the value is an empty string, but if we let this value be a dict,
        multiple plugins can add data to the dict under their own key, which avoids conflicts.

        Of course, this doesn't alleviate the problem of other plugins overwriting our data,
        but it's a convention that should probably be suggested to the INGInious devs.

        ### In practice the result is:

        ```py
        submission = {
            "_id": ...
            "grade": ...
            # ...
            "custom": {
                "coding_style": CodingStyleGrades(...) # our plugin's data
                "other_plugin": ..., # other plugin's data
            }
        }
        ```
        """

        # Check contents of "custom" key
        if not submission.get("custom"):
            submission["custom"] = {}
        elif not isinstance(submission["custom"], dict):
            submission["custom"] = {"original": submission["custom"]}
            # TODO: find out how to get id value from MongoDB ObjectId object
            self._logger.info(
                f"Stored previous custom value under submission['custom']['original'] for submission {submission}. "
                "Ensure another plugin is not overwriting custom submission data."
            )

        # Check if submission has existing coding style grades
        submission_modified = False
        g = submission["custom"].get(PLUGIN_KEY)
        try:
            # Try to parse existing grades (if any)
            grades = get_grades(g)
        except ValidationError:
            # Create new grades from enabled categories in config
            grades = get_grades(self.config.enabled)
            submission_modified = True

        # Ensure all enabled grading categories are added to the submission's custom grades
        for category in self.config.enabled.values():
            if category not in grades:
                grades.add_category(category)
                submission_modified = True

        # Finally add the CodingStyleGrade object to the submission
        submission["custom"][PLUGIN_KEY] = grades

        if submission_modified:
            self._update_submission(submission)

        # From here, we can be certain submission["custom"][PLUGIN_KEY]
        # is a CodingStyleGrades object

        return submission

    def update_submission_grades(
        self, submissionid: str, grades_data: GradesIn
    ) -> Submission:
        """Attempts to update a submission with a new set of coding style grades.

        Raises `ValidationError` if grades cannot be validated.

        Parameters
        ----------
        submissionid : `str`
            ID of the submission to update grades of.
        grades_data : `GradesIn`
            Input from grading form on the WebApp.

        Returns
        -------
        `Submission`
            Updated submission

        Raises
        -------
        `ValidationError`
            Unable to validate new grades.
        """
        # Get the submission
        submission = self.submission_manager.get_submission(
            submissionid, user_check=False
        )
        submission = self.ensure_submission_custom_key(submission)

        # Add grade data to the submission
        #
        # NOTE:
        # This is by far the worst side-effect of using Pydantic with a dynamic
        # model, and having to do partial updates to a variable number of items.
        #
        # We by necessity have to copy the grading categories from the plugin config
        # and then convert them to a dict, so we can then use dict.update()
        # with the new grades we receive from the webapp form. After updating the
        # dict, we convert it back to a CodingStyleGrades object.
        #
        # This is clunky, but it still executes very quickly though.
        #
        # We end up doing the following:
        #   * Retrieve a copy of the currently enabled categories from the config (self.config.dict()["enabled"])
        #   * Iterate through the grade data from the webapp form (grades_data)
        #   * Update each enabled grade with data from the webapp form
        #   * Add the new grades to the submission for each category (overwriting the old grades)
        #   * Update the database with the new custom grades
        conf = self.config.dict()
        grades = conf["enabled"]  # type: Dict[str, dict]
        for category in grades_data:
            if category not in grades:
                continue  # skip disabled grades
            grades[category].update(grades_data[category])

        # Validate new grades and add them to the submission
        # If validation fails, ValidationError is raised
        submission["custom"][PLUGIN_KEY] = get_grades(grades)

        return self._update_submission(submission)

    def _update_submission(self, submission: Submission) -> Submission:
        """Finds an existing submission and updates it with new data.

        Returns updated submission.
        """
        # NOTE: Optional[Submission] return type annotation? We should never be in a situation
        # where we are able to call this method on a nonexistent submission, though.

        submission = deepcopy(submission)  # copy submission before modifying
        grades = submission["custom"][PLUGIN_KEY]

        if self.config.experimental.merge_grades:
            self.merge_submission_grades(submission)

        if isinstance(grades, CodingStyleGrades):
            # Convert grades object to dict if it is a CodingStyleGrades instance
            # so that MongoDB is able to serialize the data.
            submission["custom"][PLUGIN_KEY] = grades.dict()

        return self.database.submissions.find_one_and_update(
            {"_id": submission["_id"]},
            {"$set": submission},
            return_document=ReturnDocument.AFTER,
        )

    def merge_submission_grades(self, submission: Submission) -> Submission:
        """
        ## WARNING: EXPERIMENTAL

        Finds the mean of submission base grade and coding style grade
        and updates its entry in the DB collection 'user_tasks', which is a collection
        that records top submissions for every user for every task.

        NOTE: This method has no effect if the submission that is being graded is
        _not_ the user's top submission for the given task. This is due to the way
        INGInious seems to store a sort of "meta-submission" in the 'user_tasks'
        collection, which keeps track of the user's best grade for a specific task and
        the number of submissions they have made for that task, and the ID of their
        top submission. It is that grade that is displayed on the course's task list,
        and thus it is also the one we are modifying in this method.
        Furthermore, it makes no sense to grade a submission that ISN'T the user's
        best submission, so that is also relevant!

        Submission's new grade calculation:
        `merged = (base_grade + style_mean) / 2`
        """
        # grades needs to be a CodingStyleGrades object to correctly retrieve the mean grade.
        grades = submission["custom"][PLUGIN_KEY]
        if not isinstance(grades, CodingStyleGrades):
            grades = get_grades(grades)

        base_grade = submission["grade"]
        style_mean = grades.get_mean(self.config, round_grade=False)
        mean_grade = round(base_grade + style_mean) / 2

        # Update the 'user_tasks' collection (where top submissions are stored)
        self.database.user_tasks.find_one_and_update(
            {"submissionid": submission["_id"]}, {"$set": {"grade": mean_grade}}
        )
        return submission


def submission_admin_menu(
    course: Course, task: Task, submission: Submission, template_helper: TemplateHelper
) -> str:
    return template_helper.render(
        "submission_admin_menu.html",
        template_folder=TEMPLATES_PATH,
        submission=submission,
    )


def task_list_item(
    course: Course,
    task: Task,
    tasks_data: Any,
    template_helper: TemplateHelper,
) -> str:
    """Displays a progress bar denoting the current coding style grade for a given task."""
    # TODO: optimize. Cache best submission?
    start = time.perf_counter()
    best_submission = get_best_submission(task)

    # NOTE: display best coding style grade or coding style grade of best submission?

    try:
        grades = get_grades(best_submission["custom"][PLUGIN_KEY])  # type: ignore
    except (TypeError, KeyError, ValidationError) as e:
        if isinstance(e, ValidationError):
            get_logger().warning(
                f"Cannot parse coding style grades of submission {best_submission['_id']}."
            )
        return ""  # Don't display anything

    return template_helper.render(
        "task_list_item.html",
        template_folder=TEMPLATES_PATH,
        grade=grades.get_mean(PLUGIN_CONFIG),
        submission=best_submission,
    )


def task_menu(course: Course, task: Task, template_helper: TemplateHelper) -> str:
    best_submission = get_best_submission(task)
    if best_submission is None or not has_coding_style_grades(
        best_submission, PLUGIN_KEY
    ):
        return ""

    return template_helper.render(
        "task_menu.html",
        template_folder=TEMPLATES_PATH,
        submission=best_submission,
    )


def init(
    plugin_manager: PluginManager,
    course_factory: CourseFactory,
    client: Client,
    conf: OrderedDict[str, Union[str, List[str]]],
):
    """
    Allows teachers to grade several aspect of a student submission's code style.

    Available configuration:
    ::

        plugins:
        -   plugin_module: inginious_coding_style
            name: "INGInious Coding Style"
            enabled:
                - comments
                - modularity
                - structure
                - idiomaticity
            categories:
                - id: <Category id>
                name: <Name of category>
                description: <Category description>
            experimental:
                merge_grades: false

    *name*
    Display name of the plugin

    *enabled*
    List of grading category IDs to enable. Omitting this parameter enables all default categories.

    *categories*
    Custom category definitions. Each category must contain the following:

        *id*
        Unique ID of category

        *name*
        Name of category

        *description*
        Category description

    *experimental*
    Experimental features to enable. These are not guaranteed to be robust,
    and are subject to change in future versions.

        *merge_grades*
        Recalculates a submission's grade based on the mean of automated grade
        and coding style grades. The weighting is 50% automated grade + 50% style grade average.
    """
    # Get config and make it global
    config = get_config(conf)
    global PLUGIN_CONFIG
    PLUGIN_CONFIG = config

    # Display coding style grades in list of tasks for a course
    plugin_manager.add_hook("task_list_item", task_list_item)

    # Show button to navigate to detailed coding style grades for a submission
    plugin_manager.add_hook("task_menu", task_menu)

    # Show button to navigate to coding style grading page for admins
    plugin_manager.add_hook("submission_admin_menu", submission_admin_menu)

    # Grading interface for admins
    plugin_manager.add_page(
        "/admin/codingstyle/submission/<submissionid>",
        CodingStyleGrading.as_view("codingstyleadmin", config),
    )

    # # User grade overview for a specific course  (NYI)
    # plugin_manager.add_page(
    #     "/course/<courseid>/codingstyle",
    #     CodingStyleGradesOverview.as_view("codingstyleoverview"),
    # )

    # Coding Style Grade view for a specific user submission
    plugin_manager.add_page(
        "/submission/<submissionid>/codingstyle",
        StudentSubmissionCodingStyle.as_view("codingstylesubmission", config),
    )



from functools import cached_property
from typing import Any, Dict, List, Optional, OrderedDict as OrderedDictType
from collections import OrderedDict

from ._types import Submission


def get_cache_decorator() -> Any:  # TODO: type annotation
    # try to use Python 3.9's cache decorator
    # TODO: test on python3.9
    try:
        from functools import cache  # type: ignore
    except ImportError:
        from functools import lru_cache, partial

        cache = partial(lru_cache, None)  # fall back on lru_cache(None)
    return cache


cache = get_cache_decorator()


class BestSubmission:
    def __init__(self, submissions: List[Submission]) -> None:
        # self._submissions = submissions
        self.len = len(submissions)
        best = None
        for submission in submissions:
            if best is None or submission["grade"] > best["grade"]:
                best = submission
        self.best_submission = best


class BestSubmissionCache:
    """Simple class that caches the best user submissions."""

    MAX_CACHE = 500  # TODO: add way to modify this number

    def __init__(self) -> None:
        self._cache: OrderedDictType[str, BestSubmission] = OrderedDict()

    def get(
        self, userid: str, courseid: str, submissions: List[Submission]
    ) -> Optional[Submission]:
        """Retrieves the best submission for a specific course for a given user."""
        s = self._cache.get(f"{userid}{courseid}")
        if s is not None and s.len == len(submissions):
            return s.best_submission
        return (self.add(userid, courseid, submissions)).best_submission

    def add(
        self, userid: str, courseid: str, submissions: List[Submission]
    ) -> BestSubmission:
        """Adds (and overwrites existing) user submissions for a course to the cache."""
        if len(self._cache) >= self.MAX_CACHE:
            self._cache.popitem(False)  # pop FIFO
        s = BestSubmission(submissions)
        self._cache[f"{userid}{courseid}"] = s
        return s



from typing import List, Optional, TYPE_CHECKING

from inginious.frontend.tasks import Task

from ._types import Submission, HasUserManager

if TYPE_CHECKING:
    from datetime import datetime


def get_submission_authors_realname(
    obj: HasUserManager, submission: Submission
) -> List[str]:
    """Retrieves a list of the real names of a submission's authors."""
    DEFAULT = "Unknown user"
    names = []

    # Handle missing submission authors (or invalid type of submission["username"])
    if not submission.get("username") or not isinstance(submission["username"], list):
        return [DEFAULT]

    for name in submission["username"]:
        try:
            name = obj.user_manager.get_user_realname(submission["username"][0])
        except (IndexError, KeyError):
            obj._logger.error(f"Unable to get username for submission {submission}.")
            name = DEFAULT
        names.append(name)

    return names


def get_submission_timestamp(submission: Submission) -> str:
    """Returns formatted timestamp of a submission's submission time."""
    s = submission.get("submitted_on")  # type: Optional[datetime]
    return s.strftime("%Y-%m-%d %H:%M:%S") if s else "Unknown"


def has_coding_style_grades(submission: Submission, plugin_key: str) -> bool:
    """Determines if a Submission object looks like it contains coding style grades.

    NOTE: This function only verifies that _something_ is present at `Submission["custom"][PLUGIN_KEY]`.
    The actual contents are not verified. `.grades.get_grades()` handles the
    verification of the contents.
    """
    # TODO: merge this function with get_grades or something?
    return (
        bool(submission.get("custom"))
        and isinstance(submission["custom"], dict)
        and submission["custom"].get(plugin_key)
    )


def get_best_submission(task: Task) -> Optional[Submission]:
    """Retrieves the best submission by a user for a specific task."""
    # Check if we can find any submissions at all
    submission_manager = task._plugin_manager.get_submission_manager()
    submissions = submission_manager.get_user_submissions(task)
    if not submissions:
        return None

    # We have submissions, now find the best one
    best = None
    for submission in submissions:
        if best is None or submission["grade"] > best["grade"]:
            best = submission
    return best



import logging
from functools import lru_cache


@lru_cache(None)
def get_logger() -> logging.Logger:
    # We just copy the logging style of INGInious here, because
    # plugins are loaded before the INGInious logger is configured.
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)
    logger = logging.getLogger("inginious.plugins.inginious_coding_style")
    logger.addHandler(ch)
    return logger



from pathlib import Path
from typing import Any, Dict, List, Optional, OrderedDict, Union, Tuple

from flask import redirect, request
from inginious.client.client import Client
from inginious.frontend.course_factory import CourseFactory
from inginious.frontend.courses import Course
from inginious.frontend.tasks import Task
from inginious.frontend.pages.course_admin.submission import (
    SubmissionPage,
)  # the page we want to modify
from inginious.frontend.pages.course_admin.utils import INGIniousAdminPage
from inginious.frontend.pages.utils import INGIniousAuthPage
from inginious.frontend.plugin_manager import PluginManager
from inginious.frontend.template_helper import TemplateHelper
from pymongo.collection import ReturnDocument
from bson.errors import InvalidId
from werkzeug.exceptions import NotFound, Forbidden, BadRequest
from werkzeug.wrappers.response import Response
from werkzeug.datastructures import ImmutableMultiDict
from pydantic import ValidationError

from .config import PluginConfig, get_config
from .grades import CodingStyleGrades, get_grades
from .logger import get_logger
from ._types import Submission, GradesIn

PLUGIN_PATH = Path(__file__).parent.absolute()
TEMPLATES_PATH = PLUGIN_PATH / "templates"

# The key to store style grade data in the submission's "custom" dict
PLUGIN_KEY = "coding_style"  # use plugin name from config?


class CodingStyleGradesOverview(INGIniousAuthPage):
    def GET_AUTH(self) -> str:
        """Displays all coding style grades for a given course for a user."""
        return self.user_manager.session_realname()

    def fetch_submission(self, submissionid: str) -> dict:
        """Slimmed down version of SubmissionPage.fetch_submission.
        Only returns dict (submission), instead of Tuple[Course, Task, dict]"""

        try:
            submission = self.submission_manager.get_submission(submissionid, False)
            if not submission:
                raise NotFound(description=_("This submission doesn't exist."))
        except InvalidId as ex:
            self._logger.info("Invalid ObjectId : %s", submissionid)
            raise Forbidden(description=_("Invalid ObjectId."))
        return submission


class CodingStyleGrading(INGIniousAdminPage):
    """Class that implements methods that allow administrators
    to grade the coding style of a submission."""

    _logger = get_logger()

    def __init__(self, config: PluginConfig, *args, **kwargs) -> None:
        self.config = config
        super().__init__(*args, **kwargs)

    def fetch_submission(self, submissionid: str) -> Submission:
        """Slimmed down version of `SubmissionPage.fetch_submission`.
        Only returns Submission, instead of Tuple[Course, Task, Submission]"""

        try:
            submission = self.submission_manager.get_submission(submissionid, False)
            if not submission:
                raise NotFound(description=_("This submission doesn't exist."))
        except InvalidId as ex:
            self._logger.info("Invalid ObjectId : %s", submissionid)
            raise Forbidden(description=_("Invalid ObjectId."))
        return submission

    def get_best_submission(
        self, courseid: str, taskid: str, username: str
    ) -> list:  # list what?
        return list(
            self.database.user_tasks.find(
                {"username": username, "courseid": courseid, "taskid": taskid},
                {"submissionid": 1, "_id": 0},
            )
        )

    def get_submissions(
        self, courseid: str = "tutorial", taskid: str = "04_run_student"
    ) -> List[Submission]:
        """Retrieves all submissions for a given task."""

        # get the task we want to find submissions for
        task = self.course_factory.get_task(courseid, taskid)

        # get all submissions
        submissions = self.submission_manager.get_user_submissions(
            task
        )  # type: List[Submission]

        return submissions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inginious_coding_style']

package_data = \
{'': ['*'], 'inginious_coding_style': ['templates/*']}

install_requires = \
['inginious>=0.7,<0.8', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'inginious-coding-style',
    'version': '1.0.1',
    'description': '',
    'long_description': '# INGInious Coding Style\n\nINGInious Coding Style is a plugin for INGInious that allows tutors to grade several aspect of student submissions\' coding style.\n\nINGInious Coding Style should be easy to use for both tutors and students. The plugin adds new buttons and elements to various existing menus in the application that can be used to add and view coding style grades.\n\n## Documentation\n\nFull documentation can be found here: https://pederha.github.io/inginious-coding-style/\n\n\n## Installation\n\n```bash\npip install inginious-coding-style\n```\n\n## Configuration\n\nINGInious Coding Style is highly configurable and provides granular control over which grading categories are enabled, as well as the names and descriptions of the categories.\n\nFurthermore, experimental and cutting-edge features are made available in the `experimental` section. It is not advised to enable these settings in production. When these features are production-ready, they are moved out of the `experimental` section\n\n### Minimal Configuration\n\nThe following YAML snippet provides the default plugin configuration, and is a good starting point for exploring the plugin\'s functionality:\n\n```yml\nplugins: \n-   plugin_module: inginious_coding_style\n    name: "INGInious Coding Style"\n```\n\n### Full Configuration\n\nBelow is an example of a configuration making use of all available configuration options.\n\n```yml\nplugins: \n-   plugin_module: inginious_coding_style\n    name: "INGInious Coding Style"\n    enabled:\n        # This enables all default categories + 1 custom category\n        - comments\n        - modularity\n        - structure\n        - idiomaticity\n        - coolness # Our custom category\n    categories:\n        # This is a definition for a new category\n      - id: coolness\n        name: Coolness\n        description: How cool the code looks B-)\n      # This redefines a default category\n      - id: comments\n        name: Kommentering\n        description: Hvor godt kommentert koden er.\n    experimental:\n      merge_grades: false\n```\n\n<!-- ## Known Issues -->\n\n## TODO\n\n- [ ] Add clickable link to coding style grades for each task in task list.\n<!-- - [x] Complete -->\n<!-- - [ ] Incomplete -->\n\n## Developer Notes\n\nThis plugin uses [htmx](https://htmx.org/) to provide some interactivity.\n',
    'author': 'Peder Hovdan Andresen',
    'author_email': 'pedeha@stud.ntnu.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PederHA/inginious-coding-style',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .config import PluginConfig

from typing import Dict, Optional, Union
from pydantic import BaseModel, Field, validator

from ._types import GradesIn


class Grade(BaseModel):
    grade: int = Field(default=100, ge=0, le=100)
    feedback: str = Field(default="", max_length=5000)  # prevent unbounded text input


class GradingCategory(BaseModel):
    """Represents a grading category."""

    id: str  # Key data is stored under
    name: str
    description: str
    grade: int = Field(default=100, ge=0, le=100)
    feedback: str = Field(default="", max_length=5000)  # prevent unbounded text input

    @validator("name", pre=True)
    def handle_missing_name(cls, value: Optional[str], values: Dict[str, str]) -> str:
        """Defaults name to `GradingCategory.id.title()` if its value is None."""
        if value is None:
            return values["id"].title()
        return value


class CodingStyleGrades(BaseModel):
    # https://pydantic-docs.helpmanual.io/usage/models/#custom-root-types
    __root__: Dict[str, GradingCategory] = {}

    def __bool__(self) -> bool:
        """Evaluates to true if at least one grade is defined."""
        return any(g for g in self.grades)

    def __contains__(self, other: Union[str, GradingCategory]) -> bool:
        if isinstance(other, str):
            return other in self.__root__
        elif isinstance(other, GradingCategory):
            for grade in self.__root__.values():
                if grade.id == other.id:  # We assume identical ID means they are equal
                    return True
        return False

    @property
    def grades(self) -> Dict[str, GradingCategory]:
        return self.__root__

    # TODO: rename to add_grade?
    def add_category(self, category: GradingCategory) -> None:
        """Adds a grading category

        Parameters
        ----------
        category : `GradingCategory`
            The grading category to add.
        """
        self.__root__[category.id] = category

    def remove_category(self, category: Union[str, GradingCategory]) -> None:
        """Removes a grading category.

        Parameters
        ----------
        category : Union[str, GradingCategory]
            Name of category or a `GradingCategory` object.
        """
        if isinstance(category, str):
            self.__root__.pop(category, None)
        elif isinstance(category, GradingCategory):
            self.__root__.pop(category.id, None)
        # TODO: raise exception if invalid category?
        raise TypeError(
            "Argument 'category' must be of type 'str' or 'GradingCategory'."
        )

    def get_mean(
        self,
        config: Optional[PluginConfig] = None,
        round_grade: bool = True,
        ndigits: int = 2,
    ) -> float:
        """Returns mean coding style grade.
        Optionally calculate only based on enabled grading categories.

        Parameters
        ----------
        config : `Optional[PluginConfig]`, optional
            Plugin config specifying enabled categories, by default None
        round_grade : `bool`, optional
            Whether or not to round the mean grade value, by default True
        ndigits : `int`, optional
            Rounding precision, by default 2 digits after decimal point

        Returns
        -------
        `float`
            Mean grade
        """
        if config is not None:
            grades = [v for (k, v) in self.__root__.items() if k in config.enabled]
        else:
            grades = [v for (k, v) in self.__root__.items()]
        n = len(grades) or 1  # avoid divison by 0
        avg = sum(g.grade for g in grades) / n
        if round_grade:
            return round(avg, ndigits)
        return avg

    def dict(self) -> Dict[str, GradingCategory]:  # type: ignore
        """Returns a dict version of its own `__root__` attribute."""
        return super().dict()["__root__"]


def get_grades(
    grades: Union[GradesIn, Dict[str, GradingCategory]]
) -> CodingStyleGrades:
    """Attempts to create a CodingStyleGrades object based on grade data."""
    # This function lets us change the CodingStyleGrades constructor
    # however we want without having to worry about breaking construction
    # of CodingStyleGrades objects elsewhere in the plugin, if this is the
    # canonical way to create CodingStyleGrades objects.
    return CodingStyleGrades.parse_obj(grades)


DEFAULT_CATEGORIES = {
    "comments": GradingCategory(
        id="comments",
        name="Comments",
        description="Appropriate use of comments.",
    ),
    "modularity": GradingCategory(
        id="modularity",
        name="Modularity",
        description="Modularity of the code, i.e. appropriate use of functions and encapsulation.",
    ),
    "structure": GradingCategory(
        id="structure",
        name="Structure",
        description="The quality of the code's structure, i.e. comprehensible variable names, nesting, and program flow.",
    ),
    "idiomaticity": GradingCategory(
        id="idiomaticity",
        name="Idiomaticity",
        description="How idiomatic the code is, i.e. appropriate use of language-specific constructs (list comprehensions, enumerate(), etc. for Python).",
    ),
}



from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, validator, Field
from pydantic.fields import ModelField
from .logger import get_logger
from .grades import GradingCategory, DEFAULT_CATEGORIES


class ExperimentalSettings(BaseModel):
    merge_grades: bool = False
    # automatic_grading: bool = False # NYI (obviously)
    # other experimental settings here...


class PluginConfigIn(BaseModel):
    """Maps to the plugin configuration options found in configuration.yaml"""

    # Name of plugin
    name: str = "INGInious Coding Style"

    # Enabled grading categories
    enabled: List[str] = []

    # Custom grading categories
    categories: List[GradingCategory] = []

    # Experimental settings
    experimental: ExperimentalSettings = Field(default_factory=ExperimentalSettings)

    @validator("enabled", pre=True)
    def accept_enabled_none(cls, enabled: Optional[List[str]]) -> List[str]:
        if enabled is None:
            return []
        return enabled

    @validator("categories", pre=True)
    def accept_categories_none(cls, categories: Optional[List[str]]) -> List[str]:
        if categories is None:
            return []
        return categories

    @validator("experimental", pre=True)
    def accept_experimental_none(
        cls, experimental: Optional[Dict[str, bool]], field: ModelField
    ) -> Union[ExperimentalSettings, Optional[Dict[str, bool]]]:
        if experimental is None and field.default_factory is not None:
            return field.default_factory()
        return experimental


class PluginConfig(BaseModel):
    """The config used by the plugin.

    This class consolidates the fields `enabled` and `categories` from the input config.
    Furthermore, it retrieves the corresponding `GradingCategory` for any enabled
    predfined categories.
    """

    name: str
    enabled: Dict[str, GradingCategory] = {}
    experimental: ExperimentalSettings

    def __init__(self, config_in: PluginConfigIn) -> None:
        enabled = self._make_dict_from_enabled(
            enabled=config_in.enabled,
            custom_categories={c.id: c for c in config_in.categories},
        )
        super().__init__(
            name=config_in.name,
            enabled=enabled,
            experimental=config_in.experimental,
        )

    # TODO: rename method
    def _make_dict_from_enabled(
        self, enabled: List[str], custom_categories: Dict[str, GradingCategory]
    ) -> Dict[str, GradingCategory]:
        """Fetches default categories (if any) and consolidates it with custom categories."""

        if enabled is None:
            enabled = {}

        enabled_categories: Dict[str, GradingCategory] = {}

        # Attempt to get a GradingCategory object for each enabled category
        for cat in enabled:
            # check if a custom category with this name is defined
            category = custom_categories.get(cat)
            if not category:
                # check if category is a default category
                category = DEFAULT_CATEGORIES.get(cat)
                if not category:
                    get_logger().info(
                        f"Ignoring undefined grading category '{cat}' for parameter 'enabled'."
                    )
                    continue
            # Add category
            enabled_categories[cat] = category

        if not enabled_categories:
            get_logger().warning(
                "No valid values received for option 'enabled'. "
                f"Falling back on defaults: {DEFAULT_CATEGORIES.keys()}"
            )
            return DEFAULT_CATEGORIES

        return enabled_categories


def get_config(config: Dict[str, Any]) -> PluginConfig:
    # First we validate the contents of the config file
    conf_in = PluginConfigIn(**config)
    # Then we construct the config used by the plugin
    return PluginConfig(conf_in)
